"""Utils for handling dictionaries.
"""
import inspect
from typing import Mapping, Callable


def fit_dict(dct: dict, func: Callable, accept_kwargs: bool = True) -> dict:
    """Returns a subset of (k, v) pairs from `dct` whose key
    matches an argument of the callable `func`.

    If accept_kwargs==True and the function does accept key-word arguments,
    then all arguments are accepted.
    """
    func_spec = inspect.getfullargspec(func)
    func_args = func_spec.args
    if func_spec.varkw and accept_kwargs:
        return dct

    return {
        k: v for k, v in dct.items() if k in func_args
    }


def apply_recursively(
        dct: dict,
        apply: Callable[[object], object],
        pred: Callable[[object], bool] = (lambda _k: True)
    ) -> dict:
    """Applies a function recursively to the provided dictionary, possibly
    filtering the fields to which it is applied.

    Parameters
    ----------
    dct : dict
        The dictionary to which the callable `apply` will be recursively applied.

    apply : Callable
        The function to apply to the dictionary's fields.

    pred : Callable
        Predicate to filter which fields to apply the function to.
        Receives the (key, value) pair as an input.
    """
    for k in dct.keys():
        if isinstance(dct[k], Mapping):
            dct[k] = apply_recursively(dct[k], apply, pred)
        elif pred(k):
            dct[k] = apply(dct[k])
    return dct



"""A set of functions to evaluate predictions on common performance
and fairness metrics, possibly at a specified FPR or FNR target.
"""

import math
import logging
from typing import Optional

import numpy as np
from sklearn.metrics import confusion_matrix


def safe_division(a: float, b: float):
    return 0 if b == 0 else a / b


def evaluate_performance(y_true: np.ndarray, y_pred: np.ndarray) -> dict:
    """Evaluates the provided predictions on common performance metrics.
    NOTE: currently assumes labels and predictions are binary - should we extend
    it to multi-class labels?

    Parameters
    ----------
    y_true : np.ndarray
        The true class labels.
    y_pred : np.ndarray
        The discretized predictions.

    Returns
    -------
    dict
        A dictionary with key-value pairs of (metric name, metric value).
    """
    # Compute confusion matrix
    tn, fp, fn, tp = confusion_matrix(y_true, y_pred, labels=(0, 1)).ravel()

    total = tn + fp + fn + tp
    pred_pos = tp + fp
    pred_neg = tn + fn
    assert pred_pos + pred_neg == total

    label_pos = tp + fn
    label_neg = tn + fp
    assert label_pos + label_neg == total

    results = {}
    
    # Accuracy
    results["accuracy"] = (tp + tn) / total

    # True Positive Rate (Recall)
    results["tpr"] = safe_division(tp, label_pos)

    # False Positive Rate
    results["fpr"] = safe_division(fp, label_neg)

    # True Negative Rate
    # results["tnr"] = tn / label_neg

    # Precision
    results["precision"] = safe_division(tp, pred_pos)

    # Positive Prediction Rate
    results["ppr"] = safe_division(pred_pos, total)

    return results


def evaluate_fairness(
        y_true: np.ndarray,
        y_pred: np.ndarray,
        sensitive_attribute: np.ndarray,
        return_groupwise_metrics: Optional[bool] = False,
    ) -> dict:
    """Evaluates fairness as the ratios between group-wise performance metrics.

    Parameters
    ----------
    y_true : np.ndarray
        The true class labels.
    y_pred : np.ndarray
        The discretized predictions.
    sensitive_attribute : np.ndarray
        The sensitive attribute (protected group membership).
    return_groupwise_metrics : Optional[bool], optional
        Whether to return group-wise performance metrics (bool: True) or only
        the ratios between these metrics (bool: False), by default False.

    Returns
    -------
    dict
        A dictionary with key-value pairs of (metric name, metric value).
    """
    # All unique values for the sensitive attribute
    unique_groups = np.unique(sensitive_attribute)

    results = {}
    groupwise_metrics = {}
    unique_metrics = set()

    # Helper to compute key/name of a group-wise metric
    def group_metric_name(metric_name, group_name):
        return f"{metric_name}_group={group_name}"

    assert len(unique_groups) > 1, (
        f"Found a single unique sensitive attribute: {unique_groups}")

    for s_value in unique_groups:
        # Indices of samples that belong to the current group
        group_indices = np.argwhere(sensitive_attribute == s_value).flatten()

        # Filter labels and predictions for samples of the current group
        group_labels = y_true[group_indices]
        group_preds = y_pred[group_indices]

        # Evaluate group-wise performance
        curr_group_metrics = evaluate_performance(group_labels, group_preds)

        # Add group-wise metrics to the dictionary
        groupwise_metrics.update({
            group_metric_name(metric_name, s_value): metric_value
            for metric_name, metric_value in curr_group_metrics.items()
        })

        unique_metrics = unique_metrics.union(curr_group_metrics.keys())

    # Compute ratios
    for metric_name in unique_metrics:
        ratio_name = f"{metric_name}_ratio"
        curr_metric_results = [
            groupwise_metrics[group_metric_name(metric_name, group_name)]
            for group_name in unique_groups
        ]

        results[ratio_name] = safe_division(
            min(curr_metric_results),
            max(curr_metric_results)
        )

    # Optionally, return group-wise metrics as well
    if return_groupwise_metrics:
        results.update(groupwise_metrics)

    return results


def compute_binary_predictions(
        y_true: np.ndarray,
        y_pred_scores: np.ndarray,
        threshold: Optional[float] = None,
        tpr: Optional[float] = None,
        fpr: Optional[float] = None,
        ppr: Optional[int] = None,
        random_seed: Optional[int] = 42,
    ) -> np.ndarray:
    """Discretizes the given score predictions into binary labels,
    according to the provided target metric for thresholding.

    Parameters
    ----------
    y_true : np.ndarray
        The true binary labels
    y_pred_scores : np.ndarray
        Predictions as a continuous score between 0 and 1
    threshold : Optional[float], optional
        Whether to use a specified (global) threshold, by default None
    tpr : Optional[float], optional
        Whether to target a specified TPR (true positive rate, or recall), by
        default None
    fpr : Optional[float], optional
        Whether to target a specified FPR (false positive rate), by default None
    ppr : Optional[float], optional
        Whether to target a specified PPR (positive prediction rate), by default
        None

    Returns
    -------
    np.ndarray
        The binarized predictions according to the specified target.
    """
    assert sum(1 for val in {threshold, fpr, tpr, ppr} if val is not None) == 1, (
        f"Please provide exactly one of (threshold, fpr, tpr, ppr); got "
        f"{(threshold, fpr, tpr, ppr)}."
    )

    # If threshold provided, just binarize it, no untying necessary
    if threshold:
        return (y_pred_scores >= threshold).astype(int)
    
    # Otherwise, we need to compute the allowed value for the numerator
    # and corresponding threshold (plus, may require random untying)
    label_pos = np.count_nonzero(y_true)
    label_neg = np.count_nonzero(1 - y_true)
    assert (total := label_pos + label_neg) == len(y_true)  # sanity check

    # Indices of predictions ordered by score, descending
    y_pred_sorted_indices = np.argsort(-y_pred_scores)

    # Labels ordered by descending prediction score
    y_true_sorted = y_true[y_pred_sorted_indices]

    # Number of positive predictions allowed according to the given metric
    # (the allowed budget for the metric's numerator)
    positive_preds_budget: int

    # Samples that count for the positive_preds_budget
    # (LPs for TPR, LNs for FPR, and all samples for PPR)
    # (related to the metric's denominator)
    target_samples_mask: np.ndarray
    if tpr:
        # TPs budget to ensure >= the target TPR
        positive_preds_budget = math.ceil(tpr * label_pos)
        target_samples_mask = y_true_sorted == 1  # label positive samples
    
    elif fpr:
        # FPs budget to ensure <= the target FPR
        positive_preds_budget = math.floor(fpr * label_neg)
        target_samples_mask = y_true_sorted == 0  # label negative samples

    elif ppr:
        # PPs budget to ensure <= the target PPR
        positive_preds_budget = math.floor(ppr * total)
        target_samples_mask = np.ones_like(y_true_sorted).astype(bool) # all samples

    # Indices of target samples (relevant for the target metric), ordered by descending score
    target_samples_indices = y_pred_sorted_indices[target_samples_mask]

    # Find the threshold at which the specified numerator_budget is met
    threshold_idx = target_samples_indices[positive_preds_budget]
    threshold = y_pred_scores[threshold_idx]

    ####################################
    # Code for random untying follows: #
    ####################################
    y_pred_binary = (y_pred_scores >= threshold).astype(int)

    # 1. compute actual number of positive predictions (on relevant target samples)
    actual_pos_preds = np.sum(y_pred_binary[target_samples_indices])

    # 2. check if this number corresponds to the target
    if actual_pos_preds != positive_preds_budget:
        logging.warning(
            "Target metric for thresholding could not be met, will randomly "
            "untie samples with the same predicted score to fulfill target."
        )

        assert actual_pos_preds > positive_preds_budget, (
            "Sanity check: actual number of positive predictions should always "
            "be higher or equal to the target number when following this "
            f"algorithm; got actual={actual_pos_preds}, target={positive_preds_budget};"
        )

        # 2.1. if target was not met, compute number of extra predicted positives
        extra_pos_preds = actual_pos_preds - positive_preds_budget

        # 2.2. randomly select extra_pos_preds among the relevant
        # samples (either TPs or FPs or PPs) with the same score
        rng = np.random.RandomState(random_seed)

        samples_at_target_threshold_mask = (y_pred_scores[y_pred_sorted_indices] == threshold)

        target_samples_at_target_threshold_indices = (
            y_pred_sorted_indices[
                samples_at_target_threshold_mask &      # Filter for samples at target threshold
                target_samples_mask                     # Filter for relevant (target) samples
            ]
        )

        # # The extra number of positive predictions must be fully explained by this score tie
        # import ipdb; ipdb.set_trace()   # TODO: figure out why this assertion fails
        # assert extra_pos_preds < len(target_samples_at_target_threshold_indices)

        extra_pos_preds_indices = rng.choice(
            target_samples_at_target_threshold_indices,
            size=extra_pos_preds,
            replace=False)

        # 2.3. give extra_pos_preds_indices a negative prediction
        y_pred_binary[extra_pos_preds_indices] = 0


    # Sanity check: the number of positive_preds_budget should now be exactly fulfilled
    assert np.sum(y_pred_binary[target_samples_indices]) == positive_preds_budget

    return y_pred_binary



def compute_binary_predictions_posthoc_adjustment(
        y_true: np.ndarray,
        y_pred_scores: np.ndarray,
        sensitive_attribute: np.ndarray,
        equalize_fpr: Optional[bool] = False,
        equalize_tpr: Optional[bool] = False,
        allowed_fpr_gap: Optional[float] = None,
        allowed_tpr_gap: Optional[float] = None,
        random_seed: Optional[int] = 42,
    ) -> np.ndarray:
    """Discretizes the given score predictions into binary labels, according
    to the provided fairness criteria - equalize TPR, FPR, or both.

    Parameters
    ----------
    y_true : np.ndarray
        The true binary labels.
    y_pred_scores : np.ndarray
        Predictions as a continuous score between 0 and 1.
    sensitive_attribute : np.ndarray
        Sensitive attribute representing the group each sample belongs to.
    equalize_fpr : Optional[bool], optional
        Whether to equalize group-wise FPR, by default False
    equalize_tpr : Optional[bool], optional
        Whether to equalize group-wise TPR, by default False
    allowed_fpr_gap : Optional[float], optional
        The allowed gap in group-wise FPR (if any).
    allowed_tpr_gap : Optional[float], optional
        The allowed gap in group-wise TPR (if any).
    random_seed : Optional[int], optional
        The random seed used in case some prediction randomization is required
        (e.g.,  in the case of equal odds - equalizing both TPR and FPR).

    Returns
    -------
    np.ndarray
        The binarized predictions.
    """
    assert equalize_fpr or equalize_tpr, \
        "Must target either equal FPR or equal TPR or both, got neither."

    rng = np.random.RandomState(random_seed)
    # TODO: https://github.com/AndreFCruz/hpt/issues/2




    pass



"""Computing fairness criteria on data.

File extracted from: https://github.com/fairmlbook/fairmlbook.github.io/blob/master/code/creditscore/criteria.py
"""
# flake8: noqa

import numpy as np
import pandas as pd
from scipy.spatial import Delaunay


def first_index_above(array, value):
    """Find the smallest index i for which array[i] > value.

    If no such value exists, return len(array).
    """
    array = np.array(array)
    v = np.concatenate([array > value, np.ones_like(array[-1:])])
    return np.argmax(v, axis=0)


def trisearch_int(f, lo, hi, tol=1):
    """Trinary search: minimize f(x) over [lo, hi], to within +/-tol in x.

    Works assuming f is quasiconvex.
    """
    while hi - lo > tol:
        m1 = (2*lo + hi)//3
        m2 = (lo + 2*hi)//3
        if f(m1) < f(m2):
            hi = m2-1
        else:
            lo = m1+1
    return (hi + lo)//2


def trisearch(f, lo, hi, tol):
    """Trinary search: minimize f(x) over [lo, hi], to within +/-tol in x.

    Works assuming f is quasiconvex.
    """
    while hi - lo > tol:
        m1 = (2.*lo + hi)/3
        m2 = (lo + 2.*hi)/3
        if f(m1) < f(m2):
            hi = m2
        else:
            lo = m1
    return (hi + lo)/2


class CriteriaData(object):
    """Class for computing fairness criteria of data."""

    def __init__(self, cdfs, performance, totals):
        """Create a CriteriaData instance from marginals.

        cdfs and performance are both dataframes, with index = score and columns being the groups.

        cdfs[group][score] = fraction of people of that group with score below that score.
        performance[group][score] = fraction of people with that group & score that succeed.
        totals[group] = total number of people of that group.

        totals can be either a dictionary or an array (if an array, in the same order as cdfs.columns)
        """
        self.cdfs = cdfs
        self.performance = performance
        self.columns = self.cdfs.columns
        if isinstance(totals, dict):
            totals = [totals[r] for r in self.columns]
        self.totals = np.array(totals)
        self.flipped = self.cdfs.index[0] > self.cdfs.index[-1]
        self.pdfs = self.get_pdfs()

    def get_pdfs(self):
        cdf_vs = np.concatenate([[np.zeros_like(self.cdfs.values[0])],self.cdfs.values])
        pdf_vs = (cdf_vs[1:]-cdf_vs[:-1])
        pdf = pd.DataFrame(pdf_vs, columns=self.columns, index=self.cdfs.index)
        return pdf

    @property
    def trisearch(self):
        if self.cdfs.index.dtype == 'int64':
            return trisearch_int
        else:
            return trisearch

    @classmethod
    def from_individuals(self, data, groups = None, binsize=0.025):
        """Recover CriteriaData instance from individual performances.

        data should be a dataframe with three columns: the group key, the
        predictor score, and the result.  Scores will be binned into
        bins of size binsize to compute the performance per group/score.

        groups is the set of group keys to use; it may be specified to
        restrict the set, or to define the order.
        """
        groupkey, score, result = data.columns
        if groups is None:
            groups = list(data.groupby(groupkey).mean().sort(score).index)[::-1]

        data = data.copy()
        data = data.sort(score)
        data[score] = (data[score] / binsize).astype(int)
        datasets = {}
        cdfs = pd.DataFrame(index=np.arange(min(data[score]), max(data[score])+1),
                            columns=groups)
        performance = cdfs.copy()
        cdfs = cdfs.notnull()*0
        totals = []
        for group in groups:
            grouped = data[data[groupkey] == group].groupby(score)
            num_per_score = grouped.count()[result]
            totals.append(num_per_score.sum())
            cdfs[group][num_per_score.index] = num_per_score.values
            cdfs[group] = cdfs[group].cumsum() * 1. / cdfs[group].sum()
            performance[group] = grouped.mean()
        cdfs.index = cdfs.index * binsize
        performance.index = performance.index * binsize
        performance = performance.interpolate().fillna(0)  #Remove nans
        return self(cdfs, performance, np.array(totals))


    def profit_cutoffs(self, target_rate, _=None):
        losses = np.cumsum(self.compute_area_slices(self.performance-target_rate))
        ans = {}
        for group in self.columns:
            last_bad_row = np.argmin(losses[group].values)
            try:
                cutoff = losses.index[last_bad_row + 1]
            except IndexError:
                cutoff = losses.index[-1] + (1e-9 if not self.flipped else -1e-9)
            ans[group] = cutoff
        return ans

    def compute_area_slices(self, pdfs=None, performance=None):
        if performance is None:
            performance = self.performance
        if pdfs is None:
            pdfs = self.pdfs
        perf = performance.values
        return  pdfs * perf

    def compute_area(self, cutoffs = None, performance=None):
        if performance is None:
            performance = self.performance
        if cutoffs is None:
            return self.compute_area_slices(performance=performance).sum(axis=0)
        else:
            ans = []
            for group in self.columns:
                area = self.compute_area_slices(self.pdfs[group].loc[cutoffs[group]:],
                                                performance[group].loc[cutoffs[group]:]).sum(axis=0)
                ans.append(area)
            return np.array(ans)

    def compute_curves(self):
        dfs = []
        for value in [self.performance, 1-self.performance]:
            area_slices = self.compute_area_slices(performance=value).values
            #area_slices = np.vstack([np.zeros_like(area_slices[0]), area_slices])
            under_curve = area_slices.sum(axis=0)
            fraction_excluded = area_slices.cumsum(axis=0) / under_curve
            dfs.append(pd.DataFrame(1-fraction_excluded, index=self.cdfs.index, columns=self.columns))
        return dfs

    def score_two_sided_profit(self, point, target_rate):
        profit = 0
        fraction_nondefaulters = self.compute_area()
        losses = (1 - fraction_nondefaulters) * point[1] * target_rate
        gains = fraction_nondefaulters * point[0] * (1 - target_rate)
        profit = (self.totals * (gains - losses)).sum()
        return profit

    def two_sided_optimum(self, target_rate):
        good_frac, bad_frac = self.compute_curves()
        polygons = [Delaunay(list(zip(good_frac[group], bad_frac[group]))) for group in self.columns]
        valid_points = []
        for poly in polygons:
            for p in poly.points:
                if all(poly2.find_simplex(p) != -1 for poly2 in polygons):
                    valid_points.append(p)
        valid_points = np.array(valid_points)
        result = (-1, None)
        for p in valid_points:
            score = self.score_two_sided_profit(p, target_rate)
            if score > result[0]:
                result = (score, p)
        return result[1]

    def _find_other_endpoint(self, curvex, curvey, i, p):
        if curvex[0] > curvex[-1]:
            curvex = curvex[::-1]
            curvey = curvey[::-1]
            i = len(curvex) - 1 - i
            invert = lambda j: len(curvex) - 1 - j
        else:
            invert = lambda j: j

        me = (curvex[i], curvey[i])
        if p[0] < me[0]:
            lst = range(i-1, -1, -1)
        else:
            lst = range(i+1, len(curvex), 1)
        for j in lst:
            they = (curvex[j], curvey[j])
            ratio = (p[0] - me[0]) / (they[0] - me[0])
            if ratio > 1:
                continue
            assert ratio > 0
            midy = ratio * (they[1] - me[1]) + me[1]
            if midy > p[1]:
                return invert(j)

    def two_sided_ranges(self, p):
        ans = {}
        good_frac, bad_frac = self.compute_curves()
        for group in self.columns:
            polygon = Delaunay(list(zip(good_frac[group], bad_frac[group])))
            vertices = polygon.simplices[polygon.find_simplex(p)]
            vert_to_pair = lambda v: (good_frac[group].index[v], bad_frac[group].values[v])
            vals = list(map(vert_to_pair, vertices))
            answer = (min(vals), max(vals))
            for v in vertices:
                if np.abs(p - polygon.points[v]).max() < 1e-3:
                    val = vert_to_pair(v)
                    answer = (val, val)
                    break
                j = self._find_other_endpoint(good_frac[group].values, bad_frac[group].values, v, p)
                if j is None:
                    continue

                vals = list(map(vert_to_pair, [v, j]))
                answer2 = (min(vals), max(vals))
                if abs(answer2[1][1] - answer2[0][1]) < abs(answer[1][1] - answer[0][1]):
                    answer = answer2
            ans[group] = (answer[0][0], answer[1][0])
        return ans

    def evaluate_opportunity(self, cutoffs):
        if not isinstance(cutoffs, dict):
            return np.array([cutoffs[0] for r in self.columns])
        return self.compute_area(cutoffs) / self.compute_area()

    def opportunity_cutoffs(self, target_rate, target_is_profit=False):
        if target_is_profit:
            target_rate = self.get_best_opportunity(target_rate)
        ans = {}
        cumulative = np.cumsum(self.compute_area_slices(), axis=0).values
        goal_area = cumulative[-1] * (1-target_rate)
        indices = first_index_above(cumulative, goal_area)
        cutoffs = self.cdfs.index[indices]
        return dict(zip(self.columns, cutoffs))

    def demographic_cutoffs(self, target_rate, target_is_profit=False):
        if target_is_profit:
            target_rate = self.get_best_demographic(target_rate)
        ans = {}
        for group in self.columns:
            index = first_index_above(self.cdfs[group].values, (1-target_rate))-1
            cutoff = self.cdfs.index[max(index, 0)]
            ans[group] = cutoff
        return ans

    def fixed_cutoffs(self, target_rate, target_is_profit=False):
        if target_is_profit:
            target_rate = self.get_best_fixed(target_rate)
        return {r:target_rate for r in self.columns}

    def compute_profit(self, cutoffs, target_rate):
        if isinstance(cutoffs, np.ndarray):
            # Hack, where we get the right answer for two-sided equality
            return self.score_two_sided_profit(cutoffs, target_rate)
        profits = self.compute_area(cutoffs, performance=self.performance - target_rate)
        return (profits * self.totals).sum()

    def efficiency(self, cutoffs, target_rate):
        return self.compute_profit(cutoffs, target_rate) / self.compute_profit(self.profit_cutoffs(target_rate), target_rate)

    def get_best_opportunity(self, target_rate):
        return trisearch(lambda t: -self.compute_profit(self.opportunity_cutoffs(t), target_rate), 0, 1, 1e-3)

    def get_best_fixed(self, target_rate):
        return self.trisearch(lambda t: -self.compute_profit({r:t for r in self.columns}, target_rate), self.cdfs.index.min(), self.cdfs.index.max(), 1e-3)

    def get_best_demographic(self, target_rate):
        return trisearch(lambda t: -self.compute_profit(self.demographic_cutoffs(t), target_rate), 0, 1, 1e-3)

    def coverage(self, cutoffs):
        total = 0
        covered = 0
        for i, group in enumerate(self.columns):
            if group not in cutoffs:
                continue
            total += self.totals[i]
            covered += self.totals[i] * (1 - self.cdfs[group][cutoffs[group]])
        return covered / total



"""Utils pertaining to importing and loading objects.
"""
import importlib
from typing import Union, Callable, Any


def import_object(import_path: str) -> Union[object, Callable]:
    """Imports the object at the given module/class path.

    Parameters
    ----------
    import_path : str
        The import path for the object to import.

    Returns
    -------
    The imported object (this can be a class, a callable, a variable).
    """
    separator_idx = import_path.rindex('.')
    module_path = import_path[: separator_idx]
    obj_name = import_path[separator_idx + 1:]

    module = importlib.import_module(module_path)
    return getattr(module, obj_name)


def get_full_name(obj: Any) -> str:
    """Returns identifier name for the given callable.

    Should be equal to the import path:
        obj == import_object(get_full_name(obj))

    Parameters
    ----------
    obj : object
        The object to find the classpath for.

    Returns
    -------
    The object's classpath.
    """
    if callable(obj):
        return obj.__module__ + '.' + obj.__qualname__
    else:
        return obj.__class__.__module__ + '.' + obj.__class__.__qualname__



"""Module to suggest and sample hyperparameters from distributions defined in
a hyperparameter space.
"""
import logging
from copy import deepcopy
from typing import Iterable, Union

from optuna.trial import BaseTrial

from .utils.trial import RandomValueTrial
from .utils.load_yaml import load_hyperparameter_space


def suggest_random_hyperparams(
        hyperparameter_space: Union[dict, str],
        seed: int,
    ) -> dict:
    """Suggests a random set of hyperparameters from the given hyperparameter space.
    
    NOTE: this is a deterministic function of the given seed number; the seed must
    itself be randomly drawn for each function call if you want to get a random sample
    of hyperparameter configurations.

    Parameters
    ----------
    hyperparameter_space : Union[dict, str]
        A dict or a path to a YAML file representing a hyperparameter space.
        Please pre-load the YAML file and provide it as a dict here, as otherwise
        this call will be unnecessarilly inefficient.
    seed : int
        The random seed used to generate the random set of hyperparameters.
        This function is a deterministic function of the seed provided.

    Returns
    -------
    dict
        A set of hyperparameters that was randomly drawn from the given
        hyperparameter space.
    """

    # If provided a path to a YAML file instead of a pre-loaded hyperparameter space
    if isinstance(hyperparameter_space, str):
        # Load (and validate) the given hyperparameter space
        # NOTE: you should really pre-load the YAML, doing it here will be inefficient
        hyperparameter_space = load_hyperparameter_space(hyperparameter_space)

    return suggest_callable_hyperparams(
        RandomValueTrial(seed=seed),
        hyperparameter_space,
    )


def suggest_callable_hyperparams(
        trial: BaseTrial,
        hyperparameter_space: dict,
        param_prefix: str = "learner",
    ) -> dict:
    """Suggests the top-level hyperparameters for a class instantiation, or
    for parameterizing any other callable.

    This includes the classpath/importpath, and the conditional
    hyperparameters to use as kwargs for the class.

    Parameters
    ----------
    trial : BaseTrial
        The trial object to interact with the hyperparameter sampler.
    hyperparameter_space : dict
        A dictionary representing a hyperparameter space.
    param_prefix : str
        The prefix to attach to all parameters' names in order to uniquely
        identify them.

    Returns
    -------
    Dict
        The suggested callable + its suggested key-word arguments.
    """
    assert isinstance(hyperparameter_space, dict)

    # Suggest callable type (top-level callable choice)
    callable_type = trial.suggest_categorical(
        f"{param_prefix}_type", list(hyperparameter_space.keys()),
    )

    hyperparam_subspace = hyperparameter_space[callable_type]
    callable_classpath = hyperparam_subspace['classpath']

    # Suggest callable kwargs (on conditional sub-space)
    callable_kwargs = suggest_hyperparams(
        trial, hyperparam_subspace.get('kwargs', {}),
        param_prefix=f"{param_prefix}_{callable_type}"
    )

    hyperparams = {'classpath': callable_classpath, **callable_kwargs}
    return hyperparams


def suggest_hyperparams(
        trial: BaseTrial,
        hyperparameter_space: dict,
        param_prefix: str = '',
    ) -> dict:
    """Uses the provided hyperparameter space to suggest specific
    configurations using the given Trial object.

    Parameters
    ----------
    trial : BaseTrial
        The trial object to interact with the hyperparameter sampler.
    hyperparameter_space : dict
        A dictionary representing a hyperparameter space.
    param_prefix : str
        The prefix to attach to all parameters' names in order to uniquely
        identify them.

    Returns
    -------
    An instantiation of the given hyperparameter space.
    """
    if not hyperparameter_space:
        logging.warning("Got empty hyperparameter space. Returning '{}'.")
        return dict()
    assert isinstance(hyperparameter_space, dict)

    params = dict()
    for key, value in hyperparameter_space.items():
        param_id = f'{param_prefix}_{key}'

        # Fixed value
        if isinstance(value, (str, int, float)) or not isinstance(value, Iterable):
            params[key] = value

        # Categorical parameter
        elif isinstance(value, (list, tuple)):
            # Categorical parameters must be of one of the following types
            valid_categ_types = (bool, int, float, str)

            # If not, map values to strings for optuna, then remap them back
            if not all(el is None or isinstance(el, valid_categ_types) for el in value):
                categ_map = {str(el): el for el in value}
                suggested_categ_encoded = trial.suggest_categorical(
                    param_id, sorted(categ_map.keys()),
                )
                params[key] = categ_map[str(suggested_categ_encoded)]

            # If categorical values have valid types, use them directly
            else:
                params[key] = trial.suggest_categorical(
                    param_id, value,
                )

        # Numerical parameter
        elif isinstance(value, dict) and 'type' in value:
            params[key] = suggest_numerical_hyperparam(
                trial, value, param_id,
            )

        # Nested parameter
        elif isinstance(value, dict):
            params[key] = suggest_hyperparams(
                trial, value, f"{key}_{param_prefix}",
            )

        else:
            raise ValueError(
                f"Invalid hyperparameter configuration: {key}={value} "
                f"({type(key)} -> {type(value)})."
            )

    return params


def suggest_numerical_hyperparam(
        trial: BaseTrial,
        config: dict,
        param_id: str,
    ) -> float:
    """Helper function to suggest a numerical hyperparameter.

    Parameters
    ----------
    trial : BaseTrial
        The trial object to interact with the hyperparameter sampler.
    config : dict
        The distribution to sample the parameter from.
    param_id : str
        The parameter's name.

    Returns
    -------
    The sampled value.
    """
    # Parameter's type
    config = deepcopy(config)
    param_type = config.pop('type')
    valid_param_types = ('int', 'float', 'uniform', 'discrete_uniform', 'loguniform')
    assert param_type in valid_param_types, \
        f'Invalid parameter type {param_type}, choose one of {valid_param_types}'

    # Parameter's range
    assert 'range' in config and isinstance(config['range'], (list, tuple)), \
        'Must provide a range when searching a numerical parameter'
    low, high = config.pop('range')

    # Remaining suggestion parameters are used as is (e.g., log scale)
    suggest_param_func = getattr(trial, f'suggest_{param_type}')
    return suggest_param_func(
        param_id, low, high, **config,
    )



"""A simple wrapper for optuna hyperparameter tuners.
"""

import re
import logging
import dataclasses
from pathlib import Path
from functools import partial
from inspect import signature
from typing import Callable, Optional, Union, List

import numpy as np
import optuna
from optuna import samplers
from optuna.trial import BaseTrial

from .suggest import suggest_callable_hyperparams
from .evaluation import (
    evaluate_performance,
    evaluate_fairness,
    compute_binary_predictions,
)

from .utils.api import BaseLearner
from .utils.load_yaml import load_hyperparameter_space
from .utils.classpath import import_object


class ObjectiveFunction:
    """Callable objective function to be used with optuna.
    """

    @dataclasses.dataclass
    class TrialResults:
        hyperparameters: dict
        validation_results: dict
        train_results: dict = None
        model: BaseLearner = None
        algorithm: str = None

        def __post_init__(self):
            match = re.match(
                r'^(?:.+[.])?(?P<algorithm>\w+)$',
                self.hyperparameters['classpath'])

            self.algorithm = match['algorithm']

    @staticmethod
    def evaluate_predictions(
            y_true: np.ndarray,
            y_pred_scores: np.ndarray,
            sensitive_attribute: Optional[np.ndarray] = None,
            **threshold_target,
        ) -> dict:
        """Evaluates the given predictions on both performance and fairness
        metrics (if `sensitive_attribute` is provided).

        Parameters
        ----------
        y_true : np.ndarray
            The true labels.
        y_pred_scores : np.ndarray
            The predicted scores.
        sensitive_attribute : np.ndarray
            The sensitive attribute - which protected group each sample belongs
            to.

        Returns
        -------
        dict
            A dictionary of (key, value) -> (metric_name, metric_value).
        """

        # Binarize predictions according to the given threshold target
        y_pred_binary = compute_binary_predictions(
            y_true, y_pred_scores, **threshold_target,
        )

        # Compute global performance metrics
        results = evaluate_performance(y_true, y_pred_binary)

        # (Optionally) Compute fairness metrics
        if sensitive_attribute is not None:
            results.update(evaluate_fairness(
                y_true, y_pred_binary, sensitive_attribute,
                return_groupwise_metrics=False,
            ))

        return results
    
    @staticmethod
    def instantiate_model(classpath: str, **hyperparams) -> BaseLearner:
        """Instantiates a model using the provided class path and provided
        hyperparameters.

        Parameters
        ----------
        classpath : str
            The classpath for importing the model's constructor/class.
        hyperparams : dict
            A dictionary of hyperparameter values to use as key-word arguments.

        Returns
        -------
        object
            The model object.
        """
        constructor = import_object(classpath)
        assert callable(constructor)

        # Use only arguments that are accepted by the model constructor
        # TODO: fix this, currently it's throwing away useful kwargs...
        # hyperparams_fitted = fit_dict(hyperparams, constructor)

        hyperparams_fitted = hyperparams.copy() # temporary placeholder

        if len(hyperparams) != len(hyperparams_fitted):
            logging.error(
                f'Ignoring following kwargs for model constructor: '
                f'{set(hyperparams) - set(hyperparams_fitted)}.'
            )

        return constructor(**hyperparams_fitted)
        
    @staticmethod
    def fit_model(model, X_train, y_train, s_train = None):
        sig = signature(model.fit)

        # compatibility with fairgbm
        if 'constraint_group' in sig.parameters:
            model.fit(X_train, y_train, constraint_group=s_train)

        # compatibility with fairlearn
        elif 'sensitive_features' in sig.parameters:
            model.fit(X_train, y_train, sensitive_features=s_train)
        
        # TODO: add adhoc compatibility with other libraries here

        # else, attempt to use s_train as a third positional argument
        elif s_train is not None and len(sig.parameters) > 2:
            model.fit(X_train, y_train, s_train)

        # else, train without sensitive attribute data
        else:
            if s_train is not None:
                logging.error(
                    f"Can't figure out how to use sensitive_attribute data for "
                    f"training with object of type '{type(model)}'.")

            model.fit(X_train, y_train)

        return model
    
    @property
    def results(self):
        return self._models_results

    # NOTE: I don't love this constructor API, feels cluttered
    def __init__(
            self,
            X_train,
            y_train,
            X_val,
            y_val,
            hyperparameter_space: Union[str, dict],
            eval_metric: str,
            s_train = None,
            s_val = None,
            other_eval_metric: Optional[str] = None,
            alpha: Optional[float] = 0.50,
            eval_func: Optional[Callable[..., dict]] = None,
            **threshold_target,
        ):
        self.X_train, self.y_train, self.s_train = (X_train, y_train, s_train)
        self.X_val, self.y_val, self.s_val = (X_val, y_val, s_val)

        self.hyperparameter_space = hyperparameter_space
        if isinstance(hyperparameter_space, (str, Path)):
            self.hyperparameter_space = load_hyperparameter_space(str(hyperparameter_space))

        self.eval_metric = eval_metric
        self.other_eval_metric = other_eval_metric
        self.alpha = alpha
        assert alpha is None or (0 <= alpha <= 1)
        self.eval_func = eval_func or ObjectiveFunction.evaluate_predictions
        self.eval_func = partial(self.eval_func, **threshold_target)

        # Store all results in a list as models are trained
        self._models_results: List[ObjectiveFunction.TrialResults] = list()

    def __call__(self, trial: BaseTrial) -> float:

        # Sample hyperparameters
        hyperparams = suggest_callable_hyperparams(trial, self.hyperparameter_space)

        # Construct model
        model = self.instantiate_model(**hyperparams)

        # Train model
        self.fit_model(model, self.X_train, self.y_train, self.s_train)

        # Compute predictions on validation data
        y_val_pred_scores = model.predict_proba(self.X_val)
        # TODO: computing predictions may require sensitive attribute access

        if len(y_val_pred_scores.shape) > 1:
            y_val_pred_scores = y_val_pred_scores[:, -1]

        # Evaluate validation predictions
        val_results = self.eval_func(
            self.y_val, y_val_pred_scores,
            sensitive_attribute=self.s_val,
        )

        # Store trial's results
        self._models_results.append(
            self.TrialResults(hyperparams, val_results, model=model))

        # Return scalarized evaluation metric
        if self.other_eval_metric:
            assert self.alpha is not None
            return (
                val_results[self.eval_metric] * self.alpha + 
                val_results[self.other_eval_metric] * (1 - self.alpha)
            )
        
        # Or simply a single evaluation metric value
        else:
            return val_results[self.eval_metric]
    
    def reconstruct_model(self, trial_results: TrialResults):
        model = self.instantiate_model(**trial_results.hyperparameters)
        return self.fit_model(model, self.X_train, self.y_train, self.s_train)
    
    def plot(
            self,
            x_axis: str = None,
            y_axis: str = None,
            pyplot_show: bool = True,
            **kwargs,
        ):
        x_axis = x_axis or self.eval_metric
        y_axis = y_axis or self.other_eval_metric
        if y_axis is None:
            raise TypeError(
                "No y_axis metric could be inferred as only one optimization "
                "metric was used. Please provide kwarg `y_axis`."
            )

        try:
            import seaborn as sns
            import pandas as pd
            from matplotlib import pyplot as plt

            data = pd.DataFrame([{
                'algorithm': r.algorithm,
                **r.validation_results,
            } for r in self.results])

            sns.set()
            sns.scatterplot(data, x=x_axis, y=y_axis, **kwargs)

            if pyplot_show:
                plt.show()

        except ImportError as err:
            logging.error(
                f"Necessary dependencies for plotting were not found: {err}."
                f"You can install them with `pip install hpt[plotting]`."
            )


class OptunaTuner:
    """This class is mostly useless, just a helper for common boilerplate.
    """

    def __init__(
            self,
            objective_function: Callable[[BaseTrial], float],
            sampler: Optional[samplers.BaseSampler] = None,
            seed: int = 42,
            **study_kwargs,
        ):

        self.objective_function = objective_function
        self.seed = seed
        self.study = optuna.create_study(
            sampler=(sampler or samplers.RandomSampler(self.seed)),
            **study_kwargs,
        )

    @property
    def results(self):
        return self.objective_function.results
    
    def optimize(self, **kwargs):
        return self.study.optimize(
            self.objective_function,
            **kwargs,
        )



"""An optuna Trial to output random hyperparameters outside optuna.
"""
from typing import Optional, Any
import warnings

from optuna.samplers import RandomSampler
from optuna.trial import FixedTrial
from optuna.distributions import BaseDistribution
from optuna import distributions


class RandomValueTrial(FixedTrial):
    """A Trial following optuna's API to be used by non-optuna tuners.
    Does not depend on an optuna.Study and can be used as a standalone object.
    """

    def __init__(
            self,
            number: int = 0,
            seed: Optional[int] = None,
            sampler: Optional[RandomSampler] = None
        ):
        assert not (seed and sampler), \
            f"Must provide at most one of (seed={seed}, sampler={sampler})"
        super().__init__(
            params=None,    # type: ignore
            number=number,
        )
        self.seed = seed
        self.sampler = sampler or RandomSampler(self.seed)

    def _suggest(self, name: str, distribution: BaseDistribution) -> Any:
        """Given a parameter's name and its distribution should return the
        suggested (sampled) value.
        (Template method from the Trial super-class).

        Parameters
        ----------
        name : str
            The parameter's name (so we don't suggest different values when the
            same parameter is sampled more than once).
        distribution : BaseDistribution
            The distribution to draw from.

        Returns
        -------
        The sampled value.
        """

        if name in self._distributions:
            # No need to sample if already suggested.
            distributions.check_distribution_compatibility(
                self._distributions[name], distribution,
            )
            param_value = self._suggested_params[name]

        else:
            if self._is_fixed_param(name, distribution):
                param_value = self.system_attrs["fixed_params"][name]
            elif distribution.single():
                param_value = distributions._get_single_value(distribution)
            else:
                param_value = self.sampler.sample_independent(
                    study=None, trial=self,     # type: ignore
                    param_name=name, param_distribution=distribution,
                )

        self._suggested_params[name] = param_value
        self._distributions[name] = distribution

        return self._suggested_params[name]

    def _is_fixed_param(self, name: str, distribution: BaseDistribution) -> bool:
        """Checks if the given parameter name corresponds to a fixed parameter.
        This implementation does not depend on an optuna.study.

        Parameters
        ----------
        name : str
        distribution : BaseDistribution

        Returns
        -------
        Whether the parameter is a fixed parameter.
        """
        system_attrs = self._system_attrs
        if "fixed_params" not in system_attrs:
            return False

        if name not in system_attrs["fixed_params"]:
            return False

        param_value = system_attrs["fixed_params"][name]
        param_value_in_internal_repr = distribution.to_internal_repr(param_value)

        contained = distribution._contains(param_value_in_internal_repr)
        if not contained:
            warnings.warn(
                f"Fixed parameter '{name}' with value {param_value} is out of range "
                f"for distribution {distribution}."
            )
        return contained



import re
from pathlib import Path
from setuptools import setup, find_packages


def stream_requirements(fd):
    """For a given requirements file descriptor, generate lines of
    distribution requirements, ignoring comments and chained requirement
    files.
    """
    for line in fd:
        cleaned = re.sub(r'#.*$', '', line).strip()
        if cleaned and not cleaned.startswith('-r'):
            yield cleaned


def load_requirements(txt_path):
    """Short helper for loading requirements from a .txt file.

    Parameters
    ----------
    txt_path : Path or str
        Path to the requirements file.

    Returns
    -------
    list
        List of requirements, one list element per line in the text file.
    """
    with Path(txt_path).open() as requirements_file:
        return list(stream_requirements(requirements_file))


# ---------------------------------------------------------------------------- #
#                                   Requirements                               #
# ---------------------------------------------------------------------------- #

ROOT_PATH = Path(__file__).parent
README_PATH = ROOT_PATH / 'README.md'

REQUIREMENTS_PATH = ROOT_PATH / 'requirements' / 'main.txt'
REQUIREMENTS_TEST_PATH = ROOT_PATH / 'requirements' / 'test.txt'
REQUIREMENTS_PLOTTING_PATH = ROOT_PATH / 'requirements' / 'plotting.txt'

requirements = load_requirements(REQUIREMENTS_PATH)
requirements_test = load_requirements(REQUIREMENTS_TEST_PATH)
requirements_plotting = load_requirements(REQUIREMENTS_PLOTTING_PATH)


# ---------------------------------------------------------------------------- #
#                                   Version                                    #
# ---------------------------------------------------------------------------- #
SRC_PATH = ROOT_PATH / 'src' / 'hpt'
VERSION_PATH = SRC_PATH / 'version.py'

with VERSION_PATH.open('rb') as version_file:
    exec(version_file.read())


# ---------------------------------------------------------------------------- #
#                                   SETUP                                      #
# ---------------------------------------------------------------------------- #
setup(
    name="hyperparameter-tuning",
    version=__version__,
    description="A minimal framework for running hyperparameter tuning",
    keywords=["ml", "optimization", "hyperparameter", "tuning", "fairness"],

    long_description=(README_PATH).read_text(),
    long_description_content_type="text/markdown",

    package_dir={'': 'src'},
    packages=find_packages('src', exclude=['tests', 'tests.*']),
    package_data={
        '': ['*.yaml, *.yml'],
    },
    include_package_data=True,

    python_requires='>=3.8',

    install_requires=requirements,

    extras_require={
        'plotting': requirements_plotting,
        'testing': requirements_test,
        'all': requirements_plotting + requirements_test,
        # NOTE: remember to add extra requirements to 'all' as well.
    },

    zip_safe=False,

    test_suite='tests',     # TODO: check this parameter
    tests_require=requirements_test,

    author="AndreFCruz",
    url="https://github.com/AndreFCruz/hpt",

    license='MIT',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
)



"""Utils specific to the tuners module.

Author: @sgpjesus
"""
from inspect import signature
from numbers import Number
from typing import Union

import yaml
from schema import And
from schema import Optional as Optional_
from schema import Or, Schema

from .classpath import import_object


class YamlValidator:
    def __init__(self):
        self.obs_class = None

    def assert_class_exists(self, path: str) -> bool:
        """Checks if a given module and Class exists in the current python
        environment. Saves class in global object to assert arguments.

        Parameters
        ----------
        path : str
            Classpath to the Class to be checked.

        Returns    AttributeError
            If Class does not exist within module.
        ValueError
            If classpath is malformed.
        -------
        bool
            True if check passes.

        Raises
        ------
        ModuleNotFoundError
            If module does not exist.
        AttributeError
            If Class does not exist within module.
        ValueError
            If classpath is malformed.
        """
        try:
            self.obs_class = import_object(path)

        except ModuleNotFoundError as e:
            module_name = path[: path.rindex(".")]
            raise ModuleNotFoundError(
                f"Provided module for model in YAML file does not exist. Check "
                f"for errors in '{module_name}'."
            ) from e

        except AttributeError as e:
            class_name = path[path.rindex(".") + 1:]
            raise AttributeError(
                f"Provided class for model in YAML file is not contained in "
                f"module. Check for errors in '{class_name}'."
            ) from e

        except ValueError as e:
            raise ValueError(
                f"Provided classpath is malformed. You should specify the "
                f"module and class name. Check for errors in '{path}'."
            ) from e

        return True

    def assert_argument_exists(self, argument: str) -> bool:
        """Checks if a given argument for the Model Class to be isntantiated is
        expected by the Class signature. Uses global object of model to check
        signature.

        Parameters
        ----------
        argument : str
            Argument to be checked.

        Returns
        -------
        bool
            True if check passes.

        Raises
        ------
        TypeError
            If argument is not expected in Class.
        """
        signature_params = list(signature(self.obs_class).parameters)
        if "kwargs" in signature_params or argument in signature_params:
            return True
        else:
            raise TypeError(
                f"Unexpected argument for model '{self.obs_class}':" f" '{argument}'"
            )


validator = YamlValidator()

# Schema for YAML file.
HYPERPARAMETER_SPACE_SCHEMA = Schema(
    {
        str: {  # Model name given by the user
            "classpath": And(str, validator.assert_class_exists),
            "kwargs": {
                Optional_(And(str, validator.assert_argument_exists)): Or(
                    Or(str, Number),
                    And(list, lambda x: len(x) >= 1),
                    And(
                        {
                            "type": Or(
                                "int",
                                "float",
                            ),
                            "range": And(
                                list,
                                lambda values: len(values) == 2,
                                lambda values: all(isinstance(x, Number) for x in values),
                            ),
                            Optional_("log", default=False): Or(True, False),
                        }
                    ),
                )
            },
        },
    }
)


def load_hyperparameter_space(path_or_dict: Union[str, dict]) -> dict:
    """Loads the hyperparameter space encoded as a YAML in the given path.
    If given a dict, space is already loaded and this function will return the
    same object.

    Parameters
    ----------
    path_or_dict : Union[str, dict]
        Either the path to the YAML file, or a dictionary containing a
        hyperparameter space following the expected structure.

    Returns
    -------
    The loaded hyperparameter space.
    """
    # Read hyperparameter space from the YAML file (if given)
    if isinstance(path_or_dict, str):
        try:
            with open(path_or_dict, "r") as f_in:
                hyperparameter_space = yaml.safe_load(f_in)
        except yaml.YAMLError as err:
            raise ValueError(f"{err}. Did you pass a valid YAML file ?") from err
    # Else, assume the given dictionary describes hyperparameter space
    elif isinstance(path_or_dict, dict):
        hyperparameter_space = path_or_dict
    else:
        raise ValueError(
            "Invalid value for `learner_hyperparams`. "
            "Must be either a path to a YAML file or a dict following the same structure."
        )
    # validate the configuration file
    HYPERPARAMETER_SPACE_SCHEMA.validate(hyperparameter_space)

    return hyperparameter_space



"""Collection of classes defining the API this package depends on.
"""
from abc import ABC, abstractmethod


class BaseLearner(ABC):
    """Interface for a generic learner (follows sklearn's API).
    """

    @abstractmethod
    def predict_proba(self, X):
        """Predict class probabilities for X.
        """
        raise NotImplementedError

    @abstractmethod
    def fit(self, X, y):
        """Fit learner to the provided features `X` and labels `y`.
        """
        raise NotImplementedError
